#!/usr/bin/env python

# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import signal

from argparse import ArgumentParser
from logging import getLogger, StreamHandler, Formatter, CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET
from logging.handlers import RotatingFileHandler

from prometheus_client.core import CollectorRegistry
from pysyncobj import SyncObjConf

from cockatrice import NAME, VERSION
from cockatrice.data_node import DataNode
from cockatrice.http_server import HTTPServer


def signal_handler(signal, frame):
    sys.exit(0)


signal.signal(signal.SIGINT, signal_handler)


def server_handler(args):
    # create logger and handler
    logger = getLogger(NAME)
    log_handler = StreamHandler()

    # determine log destination
    if args.log_file is not None:
        log_handler = RotatingFileHandler(args.log_file, 'a+', maxBytes=512000000, backupCount=5)

    # determine log level
    if args.log_level in ['CRITICAL', 'FATAL']:
        logger.setLevel(CRITICAL)
        log_handler.setLevel(CRITICAL)
    elif args.log_level == 'ERROR':
        logger.setLevel(ERROR)
        log_handler.setLevel(ERROR)
    elif args.log_level in ['WARNING', 'WARN']:
        logger.setLevel(WARNING)
        log_handler.setLevel(WARNING)
    elif args.log_level == 'INFO':
        logger.setLevel(INFO)
        log_handler.setLevel(INFO)
    elif args.log_level == 'DEBUG':
        logger.setLevel(DEBUG)
        log_handler.setLevel(DEBUG)
    elif args.log_level == 'NOTSET':
        logger.setLevel(NOTSET)
        log_handler.setLevel(NOTSET)
    else:
        logger.setLevel(INFO)
        log_handler.setLevel(INFO)

    # set log format
    handler_format = Formatter('%(asctime)s - %(levelname)s - %(pathname)s:%(lineno)d - %(message)s')
    log_handler.setFormatter(handler_format)

    # add log handler
    logger.addHandler(log_handler)

    # create http logger and handler
    http_logger = getLogger(NAME + '_http')
    http_log_handler = StreamHandler()

    # determine http log destination
    if args.http_log_file is not None:
        http_log_handler = RotatingFileHandler(args.http_log_file, 'a+', maxBytes=512000000, backupCount=5)

    # determine http log level
    http_logger.setLevel(INFO)
    http_log_handler.setLevel(INFO)

    # set http log format
    http_handler_format = Formatter('%(message)s')
    http_log_handler.setFormatter(http_handler_format)

    # add http log handler
    http_logger.addHandler(http_log_handler)

    # metrics registry
    metrics_registry = CollectorRegistry()

    if args.dump_file is not None:
        os.makedirs(os.path.dirname(args.dump_file), exist_ok=True)

    conf = SyncObjConf(
        fullDumpFile=args.dump_file,
        logCompactionMinTime=300,
        dynamicMembershipChange=True
    )

    data_node = DataNode(args.bind_addr, args.peer_addr, conf, args.index_dir, logger)
    http_server = HTTPServer(NAME, args.http_port, data_node, logger, http_logger, metrics_registry)
    http_server.start()


def client_handler(args):
    print(__name__)


def main():
    parser = ArgumentParser(description='cockatrice command')
    parser.add_argument('-v', '--version', action='version', version='cockatrice {}'.format(VERSION))

    subparsers = parser.add_subparsers()

    parser_server = subparsers.add_parser('server', help='see `server --help`')
    parser_server.add_argument('--http-port', dest='http_port', default=8080, metavar='HTTP_PORT', type=int,
                               help='http port')
    parser_server.add_argument('--bind-addr', dest='bind_addr', default='127.0.0.1:7070', metavar='BIND_ADDR',
                               type=str, help='host address')
    parser_server.add_argument('--peer-addr', dest='peer_addr', default=[], action='append', metavar='PEER_ADDR',
                               type=str, help='peer address')
    parser_server.add_argument('--index-dir', dest='index_dir', default='/tmp/cockatrice/index', metavar='INDEX_DIR',
                               type=str, help='index dir')
    # parser_server.add_argument('--schema-file', dest='schema_file', default=None, metavar='SCHEMA_FILE', type=str,
    #                            help='schema file')
    parser_server.add_argument('--dump-file', dest='dump_file', default='/tmp/cockatrice/raft/data.dump',
                               metavar='DUMP_FILE', type=str, help='dump file')
    parser_server.add_argument('--log-level', dest='log_level', default='DEBUG', metavar='LOG_LEVEL', type=str,
                               help='log level')
    parser_server.add_argument('--log-file', dest='log_file', default=None, metavar='LOG_FILE', type=str,
                               help='log file')
    parser_server.add_argument('--http-log-file', dest='http_log_file', default=None, metavar='HTTP_LOG_FILE', type=str,
                               help='http log file')
    parser_server.set_defaults(handler=server_handler)

    parser_client = subparsers.add_parser('client', help='see `client --help`')
    parser_client.add_argument('-H', metavar='HOST', help='host address')
    parser_client.set_defaults(handler=client_handler)

    args = parser.parse_args()
    if hasattr(args, 'handler'):
        args.handler(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time

from http import HTTPStatus
from logging import getLogger

from flask import Flask, jsonify, request, after_this_request, Response
from prometheus_client.core import CollectorRegistry, Counter, Histogram, Gauge
from prometheus_client.exposition import CONTENT_TYPE_LATEST, generate_latest

from cockatrice import NAME, VERSION
from cockatrice.schema import Schema

TRUE_STRINGS = ['true', 'yes', 'on', 't', 'y', '1']


class HTTPServer:
    def __init__(self, name, port, data_node, logger=getLogger(NAME),
                 http_logger=getLogger(NAME + '_http'), metrics_registry=CollectorRegistry()):
        self.__logger = logger
        self.__http_logger = http_logger
        self.__metrics_registry = metrics_registry

        self.__port = port
        self.__data_node = data_node

        self.__app = Flask(name)
        self.__app.add_url_rule('/', 'root', self.__root, methods=['GET'])
        self.__app.add_url_rule('/rest/<index_name>', 'get_index', self.__get_index, methods=['GET'])
        self.__app.add_url_rule('/rest/<index_name>', 'create_index', self.__create_index, methods=['PUT'])
        self.__app.add_url_rule('/rest/<index_name>', 'delete_index', self.__delete_index, methods=['DELETE'])
        self.__app.add_url_rule('/rest/<index_name>/_doc/<doc_id>', 'get_document', self.__get_document,
                                methods=['GET'])
        self.__app.add_url_rule('/rest/<index_name>/_doc/<doc_id>', 'index_document', self.__index_document,
                                methods=['PUT'])
        self.__app.add_url_rule('/rest/<index_name>/_doc/<doc_id>', 'delete_document', self.__delete_document,
                                methods=['DELETE'])
        self.__app.add_url_rule('/rest/<index_name>/_docs', 'index_documents', self.__index_documents,
                                methods=['PUT'])
        self.__app.add_url_rule('/rest/<index_name>/_docs', 'delete_documents', self.__delete_documents,
                                methods=['DELETE'])
        self.__app.add_url_rule('/rest/<index_name>/_search', 'search_documents', self.__search_documents,
                                methods=['GET', 'POST'])
        self.__app.add_url_rule('/rest/_cluster', 'cluster', self.__cluster, methods=['GET'])
        self.__app.add_url_rule('/metrics', 'metrics', self.__metrics, methods=['GET'])

        # disable Flask default logger
        self.__app.logger.disabled = True
        getLogger('werkzeug').disabled = True

        # metrics
        self.__metrics_http_requests_total = Counter(
            'http_requests_total',
            'The number of requests.',
            [
                'method',
                'endpoint',
                'status_code'
            ],
            registry=self.__metrics_registry
        )
        self.__metrics_http_requests_bytes_total = Counter(
            'http_requests_bytes_total',
            'A summary of the invocation requests bytes.',
            [
                'method',
                'endpoint'
            ],
            registry=self.__metrics_registry
        )
        self.__metrics_http_responses_bytes_total = Counter(
            'http_responses_bytes_total',
            'A summary of the invocation responses bytes.',
            [
                'method',
                'endpoint'
            ],
            registry=self.__metrics_registry
        )
        self.__metrics_http_requests_duration_seconds = Histogram(
            'http_requests_duration_seconds',
            'The invocation duration in seconds.',
            [
                'method',
                'endpoint'
            ],
            registry=self.__metrics_registry
        )
        # self.__metrics_kvs_records_count = Gauge(
        #     'kvs_records_count',
        #     'The number of kvs records.',
        #     registry=self.__metrics_registry
        # )

    def start(self):
        try:
            self.__app.run(host='0.0.0.0', port=self.__port)
        except OSError as ex:
            self.__logger.critical(ex)
        except Exception as ex:
            self.__logger.critical(ex)

    def __record_http_log(self, req, resp):
        log_message = '{0} - {1} [{2}] "{3} {4} {5}" {6} {7} "{8}" "{9}"'.format(
            req.remote_addr,
            req.remote_user if req.remote_user is not None else '-',
            time.strftime('%d/%b/%Y %H:%M:%S +0000', time.gmtime()),
            req.method,
            req.path + ('?{0}'.format(req.query_string.decode('utf-8')) if len(req.query_string) > 0 else ''),
            req.environ.get('SERVER_PROTOCOL'),
            resp.status_code,
            resp.content_length,
            req.referrer if req.referrer is not None else '-',
            req.user_agent
        )
        self.__http_logger.info(log_message)

        return

    def __record_metrics(self, start_time, req, resp):
        self.__metrics_http_requests_total.labels(
            method=req.method,
            endpoint=req.path + ('?{0}'.format(req.query_string.decode('utf-8')) if len(req.query_string) > 0 else ''),
            status_code=resp.status_code.value
        ).inc()

        self.__metrics_http_requests_bytes_total.labels(
            method=req.method,
            endpoint=req.path + ('?{0}'.format(req.query_string.decode('utf-8')) if len(req.query_string) > 0 else '')
        ).inc(req.content_length if req.content_length is not None else 0)

        self.__metrics_http_responses_bytes_total.labels(
            method=req.method,
            endpoint=req.path + ('?{0}'.format(req.query_string.decode('utf-8')) if len(req.query_string) > 0 else '')
        ).inc(resp.content_length if resp.content_length is not None else 0)

        self.__metrics_http_requests_duration_seconds.labels(
            method=req.method,
            endpoint=req.path + ('?{0}'.format(req.query_string.decode('utf-8')) if len(req.query_string) > 0 else '')
        ).observe(time.time() - start_time)

        # self.metrics_kvs_records_count.set(self.data_node.len())

        return

    def __post_process(self, start_time, req, resp):
        self.__record_http_log(req, resp)
        self.__record_metrics(start_time, req, resp)

        return resp

    def __root(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        resp = Response()
        resp.data = NAME + ' ' + VERSION + ' is running.'
        resp.status_code = HTTPStatus.OK
        resp.content_type = 'text/plain; charset="UTF-8"'

        return resp

    def __get_index(self, index_name):
        start_time = time.time()

        data = {}
        status_code = None

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        try:
            index_data = {'name': index_name}

            index = self.__data_node.get_index(index_name)
            index_data['doc_count'] = index.doc_count()
            index_data['doc_count_all'] = index.doc_count_all()
            index_data['last_modified'] = index.last_modified()
            index_data['latest_generation'] = index.latest_generation()
            index_data['version'] = index.version
            index_data['storage'] = {
                'folder': index.storage.folder,
                'supports_mmap': index.storage.supports_mmap,
                'readonly': index.storage.readonly,
                'files': index.storage.list()
            }

            data['index'] = index_data

            status_code = HTTPStatus.OK
        except KeyError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.NOT_FOUND
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        resp = jsonify(data)
        resp.status_code = status_code

        return resp

    def __create_index(self, index_name):
        start_time = time.time()

        data = {}
        status_code = None

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            use_ram_storage = False
            if request.args.get('use_ram_storage', default='', type=str).lower() in TRUE_STRINGS:
                use_ram_storage = True

            schema = Schema(request.data)

            self.__data_node.create_index(index_name, schema, use_ram_storage=use_ram_storage, sync=sync)

            if sync:
                status_code = HTTPStatus.CREATED
            else:
                status_code = HTTPStatus.ACCEPTED
        except json.decoder.JSONDecodeError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.BAD_REQUEST
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        resp = jsonify(data)
        resp.status_code = status_code

        return resp

    def __delete_index(self, index_name):
        start_time = time.time()

        data = {}
        status_code = None

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            self.__data_node.delete_index(index_name, sync=sync)

            if sync:
                status_code = HTTPStatus.OK
            else:
                status_code = HTTPStatus.ACCEPTED
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        resp = jsonify(data)
        resp.status_code = status_code

        return resp

    def __get_document(self, index_name, doc_id):
        start_time = time.time()

        data = {}
        status_code = None

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        try:
            results_page = self.__data_node.get_document(index_name, doc_id)

            if results_page.total <= 0:
                raise KeyError('{0} does not exist in {1}'.format(doc_id, index_name))

            fields = {}

            hit = results_page.results[0]
            for i in hit.iteritems():
                fields[i[0]] = i[1]

            doc = {'fields': fields}
            data['doc'] = doc

            status_code = HTTPStatus.OK
        except KeyError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.NOT_FOUND
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        resp = jsonify(data)
        resp.status_code = status_code

        return resp

    def __index_document(self, index_name, doc_id):
        start_time = time.time()

        data = {}
        status_code = None

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            self.__data_node.index_document(index_name, doc_id, json.loads(request.data, encoding='utf-8'), sync=sync)

            if sync:
                status_code = HTTPStatus.CREATED
            else:
                status_code = HTTPStatus.ACCEPTED
        except json.decoder.JSONDecodeError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.BAD_REQUEST
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        resp = jsonify(data)
        resp.status_code = status_code

        return resp

    def __delete_document(self, index_name, doc_id):
        start_time = time.time()

        data = {}
        status_code = None

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            self.__data_node.delete_document(index_name, doc_id, sync=sync)

            if sync:
                status_code = HTTPStatus.OK
            else:
                status_code = HTTPStatus.ACCEPTED
        except KeyError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.NOT_FOUND
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        resp = jsonify(data)
        resp.status_code = status_code

        return resp

    def __index_documents(self, index_name):
        start_time = time.time()

        data = {}
        status_code = None

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            cnt = self.__data_node.index_documents(index_name, json.loads(request.data), sync=sync)
            if cnt is not None:
                data['count'] = cnt

            if sync:
                status_code = HTTPStatus.CREATED
            else:
                status_code = HTTPStatus.ACCEPTED
        except json.decoder.JSONDecodeError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.BAD_REQUEST
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        resp = jsonify(data)
        resp.status_code = status_code

        return resp

    def __delete_documents(self, index_name):
        start_time = time.time()

        data = {}
        status_code = None

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            cnt = self.__data_node.delete_documents(index_name, json.loads(request.data), sync=sync)
            if cnt is not None:
                data['count'] = cnt

            if sync:
                status_code = HTTPStatus.OK
            else:
                status_code = HTTPStatus.ACCEPTED
        except KeyError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.NOT_FOUND
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        resp = jsonify(data)
        resp.status_code = status_code

        return resp

    def __search_documents(self, index_name):
        start_time = time.time()

        data = {}
        results = {}
        status_code = None

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        try:
            query = request.args.get('query', default='', type=str)

            search_field = request.args.get('search_field', default=self.__data_node.get_index(
                index_name).schema.get_default_search_field(), type=str)
            page_num = request.args.get('page_num', default=1, type=int)
            page_len = request.args.get('page_len', default=10, type=int)

            results_page = self.__data_node.search_documents(index_name, query, search_field, page_num,
                                                             page_len=page_len)

            results['is_last_page'] = results_page.is_last_page()
            results['page_count'] = results_page.pagecount
            results['page_len'] = results_page.pagelen
            results['page_num'] = results_page.pagenum
            results['total'] = results_page.total

            if results_page.pagecount >= page_num or results_page.total <= 0:
                hits = []
                for hit in results_page.results[results_page.offset:]:
                    doc = {'fields': {}}
                    for item in hit.iteritems():
                        doc['fields'][item[0]] = item[1]

                    h = {
                        'doc': doc,
                        'score': hit.score,
                        'rank': hit.rank,
                        'pos': hit.pos
                    }

                    hits.append(h)
                results['hits'] = hits
            else:
                raise ValueError('page_num must be <= {0}'.format(results_page.pagecount))

            data['results'] = results

            status_code = HTTPStatus.OK
        except ValueError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.MISDIRECTED_REQUEST
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        resp = jsonify(data)
        resp.status_code = status_code

        return resp

    def __cluster(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        data = {}
        status_code = None

        try:
            data['cluster_status'] = self.__data_node.getStatus()

            status_code = HTTPStatus.OK
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        resp = jsonify(data)
        resp.status_code = status_code

        return resp

    def __metrics(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            return self.__post_process(start_time, request, response)

        resp = Response()
        status_code = None

        try:
            resp.data = generate_latest(self.__metrics_registry)
            status_code = HTTPStatus.OK
        except Exception as ex:
            resp.data = '{0}\n{1}'.format(status_code.phrase, status_code.description)
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)

        resp.status_code = status_code
        resp.content_type = CONTENT_TYPE_LATEST

        return resp

    def get_test_client(self, ues_cookies=True, **kwargs):
        return self.__app.test_client(ues_cookies, **kwargs)



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from logging import getLogger

from pysyncobj import SyncObj, replicated
from whoosh.filedb.filestore import FileStorage, RamStorage
from whoosh.qparser import QueryParser


class DataNode(SyncObj):
    def __init__(self, bind_addr, peer_addrs, conf, index_dir, logger=getLogger(__name__)):
        super(DataNode, self).__init__(bind_addr, peer_addrs, conf=conf)

        self.__logger = logger

        self.index_dir = index_dir
        os.makedirs(self.index_dir, exist_ok=True)

        self.indices = {}
        self.ram_storage = RamStorage()
        self.file_storage = FileStorage(self.index_dir, supports_mmap=True, readonly=False, debug=False)

    def destroy(self):
        for index in self.indices.values():
            index.close()

        super().destroy()

    @replicated
    def create_index(self, index_name, schema, use_ram_storage=False):
        if use_ram_storage:
            if self.ram_storage.index_exists(indexname=index_name):
                self.indices[index_name] = self.ram_storage.open_index(indexname=index_name, schema=schema)
                self.__logger.info('open {0} on memory'.format(index_name))
            else:
                self.indices[index_name] = self.ram_storage.create_index(schema, indexname=index_name)
                self.__logger.info('create {0} on memory'.format(index_name))
        else:
            if self.file_storage.index_exists(indexname=index_name):
                self.indices[index_name] = self.file_storage.open_index(indexname=index_name, schema=schema)
                self.__logger.info('open {0} on {1}'.format(index_name, self.index_dir))
            else:
                self.indices[index_name] = self.file_storage.create_index(schema, indexname=index_name)
                self.__logger.info('create {0} on {1}'.format(index_name, self.index_dir))
        return self.indices[index_name]

    @replicated
    def delete_index(self, index_name):
        __index = self.indices.pop(index_name, None)
        if __index is not None:
            # # clear index
            # __index.writer().commit(mergetype=CLEAR)
            # self.__logger.info('delete {0} on {1}'.format(index_name, self.index_dir))

            # close index
            __index.close()
            self.__logger.info('close {0} on {1}'.format(index_name, self.index_dir))

            # delete files
            prefix = "_%s_" % index_name
            for filename in __index.storage:
                if filename.startswith(prefix):
                    __index.storage.delete_file(filename)
                    self.__logger.info('delete {0} on {1}'.format(index_name, os.path.join(self.index_dir, filename)))

    def index_exists(self, index_name):
        if index_name in self.indices:
            return True
        else:
            return False

    def get_index(self, index_name):
        return self.indices[index_name]

    @replicated
    def index_document(self, index_name, doc_id, fields):
        writer = self.indices[index_name].writer()

        try:
            doc = {self.indices[index_name].schema.get_unique_field(): doc_id}
            doc.update(fields)

            writer.update_document(**doc)
            writer.commit()
        except Exception as ex:
            self.__logger.error(ex)
            writer.cancel()

    @replicated
    def delete_document(self, index_name, doc_id):
        writer = self.indices[index_name].writer()

        try:
            writer.delete_by_term(self.indices[index_name].schema.get_unique_field(), doc_id)
            writer.commit()
        except Exception as ex:
            self.__logger.error(ex)
            writer.cancel()

    @replicated
    def index_documents(self, index_name, docs):
        writer = self.indices[index_name].writer()

        cnt = 0
        try:
            for doc in docs:
                writer.update_document(**doc)
                cnt = cnt + 1

            writer.commit()
        except Exception as ex:
            self.__logger.error(ex)
            writer.cancel()
            cnt = 0

        return cnt

    @replicated
    def delete_documents(self, index_name, doc_ids):
        writer = self.indices[index_name].writer()

        cnt = 0
        try:
            for doc_id in doc_ids:
                writer.delete_by_term(self.indices[index_name].schema.get_unique_field(), doc_id)
                cnt = cnt + 1

            writer.commit()
        except Exception as ex:
            self.__logger.error(ex)
            writer.cancel()
            cnt = 0

        return cnt

    def get_document(self, index_name, doc_id):
        return self.search_documents(index_name, doc_id, self.indices[index_name].schema.get_unique_field(), 1, 1)

    def search_documents(self, index_name, query, search_field, page_num, page_len=10, **kwargs):
        searcher = self.indices[index_name].searcher()

        query_parser = QueryParser(search_field, self.indices[index_name].schema)
        query = query_parser.parse(query)

        results_page = searcher.search_page(query, page_num, pagelen=page_len, **kwargs)

        return results_page



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from copy import deepcopy

import yaml
from whoosh.fields import Schema as WhooshSchema


def get_instance(class_name, **class_args):
    class_data = class_name.split('.')

    module_path = '.'.join(class_data[:-1])
    class_name = class_data[-1]

    module = importlib.import_module(module_path)
    class_obj = getattr(module, class_name)

    if class_args:
        return class_obj(**class_args)
    else:
        return class_obj()


class Schema(WhooshSchema):
    def __init__(self, scheme_yaml):
        super().__init__()

        self.__dict = {}

        try:
            self.__dict = yaml.safe_load(scheme_yaml)

            for field_name in self.__dict['schema'].keys():
                field_type = self.__get_field_type(self.__dict['schema'][field_name]['field_type'])
                for arg in self.__dict['schema'][field_name]['args'].keys():
                    setattr(field_type, arg, self.__dict['schema'][field_name]['args'][arg])
                self.add(field_name, field_type, glob=False)
        except Exception as ex:
            raise ex

    def __get_filter(self, name):
        class_name = self.__dict['filters'][name]['class']
        class_args = {}
        if 'args' in self.__dict['filters'][name]:
            class_args = deepcopy(self.__dict['filters'][name]['args'])

        instance = get_instance(class_name, **class_args)

        return instance

    def __get_tokenizer(self, name):
        class_name = self.__dict['tokenizers'][name]['class']
        class_args = {}
        if 'args' in self.__dict['tokenizers'][name]:
            class_args = deepcopy(self.__dict['tokenizers'][name]['args'])

        instance = get_instance(class_name, **class_args)

        return instance

    def __get_analyzer(self, name):
        instance = None

        if 'class' in self.__dict['analyzers'][name]:
            class_name = self.__dict['analyzers'][name]['class']
            class_args = {}
            if 'args' in self.__dict['analyzers'][name]:
                class_args = deepcopy(self.__dict['analyzers'][name]['args'])

            instance = get_instance(class_name, **class_args)
        elif 'tokenizer' in self.__dict['analyzers'][name]:
            instance = self.__get_tokenizer(self.__dict['analyzers'][name]['tokenizer'])
            if 'filters' in self.__dict['analyzers'][name]:
                for filter_name in self.__dict['analyzers'][name]['filters']:
                    instance = instance | self.__get_filter(filter_name)

        return instance

    def __get_field_type(self, name):
        class_name = self.__dict['field_types'][name]['class']
        class_args = {}
        if 'args' in self.__dict['field_types'][name]:
            class_args = deepcopy(self.__dict['field_types'][name]['args'])
            if 'analyzer' in class_args:
                class_args['analyzer'] = self.__get_analyzer(class_args['analyzer']) if class_args['analyzer'] else None
            if 'tokenizer' in class_args:
                class_args['tokenizer'] = self.__get_tokenizer(class_args['tokenizer']) if class_args[
                    'tokenizer'] else None

        instance = get_instance(class_name, **class_args)

        return instance

    def get_unique_field(self):
        for name, obj in self.items():
            if obj.unique:
                return name

        return None

    def get_default_search_field(self):
        return self.__dict['default_search_field']



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup

import os

from cockatrice import NAME, VERSION

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)),  'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=NAME,
    version=VERSION,
    description='Full text search and indexing server.',
    long_description=long_description,
    author='Minoru Osuka',
    author_email='minoru.osuka@gmail.com',
    license='AL2',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Natural Language :: Japanese',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Database',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development',
        'Topic :: System',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Clustering',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Indexing'
    ],
    url='https://github.com/mosuka/cockatrice',
    packages=[
        'cockatrice'
    ],
    platforms='any',
    install_requires=[
        'pysyncobj==0.3.4',
        'flask==1.0.2',
        'prometheus_client==0.3.1',
        'whoosh==2.7.4',
        'pyyaml==3.13',
        'janome==0.3.6'
    ],
    entry_points={
        'console_scripts': [
            'cockatrice = cockatrice.__main__:main'
        ]
    },
    test_suite='tests'
)
